/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  Seldom-used routines (overlay 1) -- mainly outside/external            *
 *  program stuff.                                                         *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <mem.h>
#include <time.h>
#include <io.h>
#include <dos.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <process.h>
#include <errno.h>
#include "prog.h"
#include "alc.h"
#include "max.h"
#include "max_msg.h"
#include "dr.h"

static int pascal Form_Outside_Filename(char *in,char *out,int method);
static void pascal Outside_Reread(int tr);



void pascal Shell_To_Dos()
{
  char *comspec;

  if (no_shell)
    return;

  if (*PRM(shelltodos))
    Display_File(0,NULL,PRM(shelltodos));

  comspec=getenv("COMSPEC");
  comspec=comspec ? comspec : "COMMAND.COM";

  display_line=1;
  Puts(WHITE);

  Outside(NULL,NULL,OUTSIDE_run,comspec,FALSE,CTL_none,RESTART_menu,NULL);

  if (*PRM(backfromdos))
    Display_File(0,NULL,PRM(backfromdos));

  Printf(attr_string,mdm_attr);  /* Set us back to the right colour */

  if (inmagnet)
    Fix_MagnEt();
}



/* The routine which controls all outside programs.  Leaving and
   returning are ASCIIZ pointers to filenames which are displayed
   before and after running an outside program, respectively.  If
   either (or both) of the filenames are NULL, nothing will be
   displayed.  Method is the method in which the program is invoked
   (dos, run, exit with erl, etc.), and the definitions can be found
   in MAX.H as the OUTSIDE_xxxx #defines. Parm is either the
   errorlevel to exit with (if method==OUTSIDE_errorlevel), or the
   ASCIIZ string of parameters to invoke the program with, if
   method==OUTSIDE_dos or OUTSIDE_run. Display_slogan determines if we
   print our logo ("Maximus-CBCS vx.yy" (etc)) upon returning from the
   program (and before the returning file is displayed.)  ctltype
   determines if we write a <progname>.CTL file before we invoke the
   external program.  If ctltype==CTL_none, then no .CTL is written.
   If it is CTL_normal, we write a normal-format .CTL file.  If it is
   CTL_upload or CTL_download, we include extra file-transfer information.
   We always check for a log fil if ctltype != CTL_none.  However, if the log
   file cannot be found, an error message will only be printed if ctltype==
   CTL_upload or CTL_download.                                              */

int pascal Outside(char *leaving,char *returning,int method,char *parm,
                   int display_slogan,int ctltype,char restart_type,
                   char *restart_name)
{
#ifdef ORACLE
  Printf(WHITE ">>> External program: `%s' <<<  ",parm);
  Press_ENTER();
  return 0;
#else

  struct _css css;

  FILE *bat;

  char temp[PATHLEN],
       temp2[PATHLEN],
       *args[MAX_EXECARGS],
       *p,
       *s;

  int erl,
      stay,
      reread,
      disk,
      max_d,
#ifndef OS_2
      save_stdout,
      save_stdin,
#endif
      timeremain,
      usrtime,
      x;

  byte save_inchat;
  
  long starttime;

  ChatSaveStatus(&css);
  ChatSetStatus(FALSE,s(cs_outside));

  erl=-10;
  save_inchat=inchat;
  inchat=FALSE;

  /* Copy the stay/reread portion of the method flag into temporary        *
   * variables.  Other parts of this function expect method to be a        *
   * scalar integer (and not a bitfield), so we have to do it this way.    */

  stay=reread=FALSE;

  if (method & OUTSIDE_stay)
    stay=TRUE;

  if (method & OUTSIDE_reread)
    reread=TRUE;


  /* Now turn off the bits, after we've copied them out */

  method &= ~(OUTSIDE_stay | OUTSIDE_reread);

  if (leaving)
    Display_File(0,NULL,leaving);

  /*************************************************************************/
  /*                  Fix lastread pointer, if in message area             */
  /*************************************************************************/

  if (method != OUTSIDE_errorlevel && inmsgarea)
    Fix_Lastread();

  /*************************************************************************/
  /*                    Write LASTUSER and RESTART files                   */
  /*************************************************************************/

  timeremain=timeleft();
  Write_LastUser();

  starttime=time(NULL);

  if (method==OUTSIDE_errorlevel)
    Write_Restart(restart_type,restart_name,rst_offset,usr.time,ctltype,
                  parm,starttime,&css);

  Mdm_flush();
  mdm_deinit();


  /*************************************************************************/
  /*                       Save Directories (if allowed)                   */
  /*************************************************************************/

  /* Save the current active drive */
  disk=getdisk();

  /* Find out how many drives we have available by calling setdisk().    *
   * This call doesn't actually CHANGE anything, but we need the return  *
   * value to figure out which drives to save.                           */

  max_d=setdisk(disk);
  max_d=min(max_d,MAX_DRIVES);

  /* Scan the "Save Directories" statement to see if we were supposed to *
   * save any drives...                                                  */

  for (x=0;x < max_d;x++)
    if (IsBit(prm.drives_to_save,x))
      break;

  /* If we are saving some drives, go into this next loop... */

  if (x != max_d)
  {
    /* If the current disk is less than 2 (ie. A: or B:), then we should   *
     * assume that we're running a floppy-based system, and should save    *
     * all of the drives we can get.  Otherwise, we're probably running    *
     * on a HARD DISK, and should only save the drives above A: and B:.    *
     * We also save as many drives as we can get our hands on...           */

    for (x=0;x < max_d;x++)
    {
      if (IsBit(prm.drives_to_save,x))
      {
        sprintf(temp,"%c:.",'A'+x);
        Save_Directory3(temp);

        if (!stay)
        {
          if (orig_path2[x])
            chdir(orig_path2[x]);

          if (orig_path[x])
            chdir(orig_path[x]);
        }
      }
    }

    /* Change us back to the right disk drive */

    setdisk(orig_disk==-1 ? disk : orig_disk);
  }



  /*************************************************************************/
  /*                       Convert command characters                      */
  /*************************************************************************/

  if (strchr(parm,'%'))
    Parse_Outside_Cmd(parm,temp2);
  else strcpy(temp2,parm);

  parm=temp2;

  /* Inside the program, usr.time means time of previous calls today.   *
   * However, external programs use this to mean the number of minutes  *
   * for today INCLUDING the current call, so we have to add in the     *
   * current time on-line before going outside, and subtract it after.  */

  usrtime=usr.time; /* Save original value */
  usr.time += timeonline();



  /*************************************************************************/
  /*                       Write External Control File                     */
  /*************************************************************************/

  if (ctltype != CTL_none)
    Write_External_Ctlfile(ctltype,parm,method);


  /*************************************************************************/
  /*                     Disable DESQview Shadoe Buffer                    */
  /*************************************************************************/

  if (displaymode==VIDEO_ibm)
  {
    Draw_StatusLine(STATUS_remove);

    if (multitasker==MULTITASKER_desqview)
      End_Shadow();
  }

#ifdef NEVER    /* This section of code was causing all sorts of probs! */

  /* Now add our tag to the DOS prompt, so SysOp doesn't forget we're      *
   * shelled out.                                                          */

  sprintf(new_prompt,"PROMPT=[%s v%s Shell]$_%s",
          us,version,original_prompt ? original_prompt : "$n$g");

  putenv(new_prompt);

#endif

  /*************************************************************************/
  /*                       Disable the ^C/^brk handler                     */
  /*************************************************************************/

  /* Temporarily get rid of our ^C handler, in case that we get SWAPPED to *
   * disk by some other program.  If we DIDN'T do this, then the computer  *
   * would branch to V'ger if the Sysop hit ^C or ^break while we were     *
   * away...  Chain will uninstall the ^C handler in a different place,    *
   * which is why we don't try to do it here.                              */

  if (did_trap && method != OUTSIDE_chain)
  {
    /* Print something to stdio before removing the keyboard handler --    *
     * DOS doesn't like to execute the 0x23 interrupt until something      *
     * is printed, and we want to make sure that if it is going to         *
     * happen, that it does so while we have our handlers in place.        */

    fprintf(stdout,"\x08");

    brkuntrap();
  }

  /* Put cursor back to real position, and flush output */
  vbuf_flush();

  #ifndef OS_2
  save_stdout=dup(fileno(stdout));
  save_stdin=dup(fileno(stdin));
  #endif

  switch (method)
  {

  /*************************************************************************/
  /*                            ERRORLEVEL EXIT                            */
  /*************************************************************************/

    case OUTSIDE_errorlevel:                   /* ErrorLevel */
      erl=atoi(parm);

      if (erl==255 || erl <= 5)
        logit(s(log_bad_erl));
      else
      {
        sprintf(temp,s(erl_xx),parm);
        logit(external_prog,temp);

        if ((p=firstchar(parm,ctl_delim,2)) != NULL && *p)
        {
          if (task_num)
            sprintf(temp,"%sERRORL%02x.BAT",original_path,task_num);
          else sprintf(temp,"%sERRORLVL.BAT",original_path);

          if ((bat=shfopen(temp,fopen_write,O_WRONLY | O_CREAT | O_TRUNC))==NULL)
            logit(cantopen,temp);
          else
          {
            fprintf(bat,"%s\n",p);
            fclose(bat);
          }
        }

        if ((prm.flags & FLAG_watchdog) && !local)
          mdm_watchdog(1);

        nowrite_lastuser=TRUE;
        quit(erl);

        /* In case that call somehow bombs... */
        rst_offset=-1L;
      }
      break;

  /*************************************************************************/
  /*                          RUN/CHAIN EXIT                               */
  /*************************************************************************/

    case OUTSIDE_run:                   /* RUN */
    case OUTSIDE_chain:                 /* Same as RUN, but we don't get control after */
    case OUTSIDE_concur:

      strcpy(temp,parm);

      logit(external_prog,parm);

      p=strtok(temp,ctl_delim);

      for (x=0;p != NULL && x < MAX_EXECARGS;x++)
      {
        switch (*p)
        {
          /* For handling redirection... */

          case '<':
            freopen(p+1,fopen_read,stdin);
            x--;
            break;

          case '>':
            if (*(p+1)=='>')  /* Is it for append? */
              freopen(p+2,fopen_append,stdout);
            else freopen(p+1,fopen_write,stdout);

            x--;
            break;

          default:
            if ((args[x]=(char *)malloc(strlen(p)+1))==NULL)
            {
              logit("!OA-MEMOVFL");
              erl=-1;
              goto RetProc;
            }

            strcpy(args[x],p);

            s=strrchr(args[x],'.');

            if (x==0 && (s==NULL || !(strnicmp(s,".exe",4)==0 ||
                                      strnicmp(s,".com",4)==0)))
            {
              logit(s(log_badex1));
              logit(s(log_badex2));

              args[++x]=NULL;
              erl=-1;
              goto Skip;
            }
            break;
        }

        p=strtok(NULL,ctl_delim);
      }

      args[x]=NULL;

      Close_Files();

      if (method==OUTSIDE_chain)
      {
        /* Close files, terminate keyboard handler, etc. */

        nowrite_lastuser=TRUE;
        FinishUp2(FALSE);
        nowrite_lastuser=FALSE;

        if (did_trap)
          brkuntrap();
      }

      if ((prm.flags & FLAG_watchdog) && !local)
        mdm_watchdog(1);

      IoPause();

      erl=spawnvp(method==OUTSIDE_chain ? P_OVERLAY : P_WAIT,args[0],args);

      IoResume();

      if ((prm.flags & FLAG_watchdog) && !local)
        mdm_watchdog(0);

#ifndef OS_2

      dup2(save_stdin,fileno(stdin));
      dup2(save_stdout,fileno(stdout));

      close(save_stdin);
      close(save_stdout);

#endif

      Reopen_Files();

      if (erl==-1)
      {
        switch(errno)
        {
          case ENOENT:
            logit(s(log_badnf));
            break;

          case ENOMEM:
            logit(s(log_badnm));
            break;

          default:
            logit(s(log_badee));
            break;
        }
      }

Skip:

      for (x=0;args[x];x++)
        sfree(args[x]);

      strcpy(temp,parm);

      logit(return_prog,parm,erl);
      break;

  /*************************************************************************/
  /*                                DOS EXIT                               */
  /*************************************************************************/

    case OUTSIDE_dos:                   /* DOS */
      strcpy(temp,parm);

      logit(external_prog,temp);

      Close_Files();

      if ((prm.flags & FLAG_watchdog) && !local)
        mdm_watchdog(1);

      IoPause();

      erl=system(temp);

      IoResume();

      if ((prm.flags & FLAG_watchdog) && !local)
        mdm_watchdog(0);

      Reopen_Files();

      logit(return_prog,temp,erl);
      break;

    default:
      logit(s(log_badom));
      break;
  }

RetProc:

  if (reread)
    Outside_Reread(timeremain);

  /*************************************************************************/
  /*                Restart Flow Control and ^brk handler                  */
  /*************************************************************************/

  if (did_trap)
    brktrap();

  Mdm_Flow_On();


  /*************************************************************************/
  /*                     Restart DESQview Shadow Buffer                    */
  /*************************************************************************/

  if (displaymode==VIDEO_ibm && multitasker==MULTITASKER_desqview)
    Start_Shadow();

#ifdef NEVER    /* This section of code was causing all sorts of probs! */

  /* Now restore our prompt string...  (Not as if it does any good, since  *
   * this is just our copy of the environment, but it'll be more           *
   * compatible with other(tm> systems!)                                   */

  sprintf(new_prompt,"PROMPT=%s",original_prompt);
  putenv(new_prompt);

#endif

  /*************************************************************************/
  /*                       Reset Current Directories                       */
  /*************************************************************************/

  Restore_Directories3();

  /* Reset More [Y,n,=]? prompts, too! */

  display_line=display_col=1;

  usr.time=usrtime;

  /*************************************************************************/
  /*                        Resinstall the FOSSIL                          */
  /*************************************************************************/

  /* The next reinit is just so we can make sure the FOSSIL is still *
   * there, and that an external program didn't borrow our interrupt *
   * vector, and forget to give it back.  (How rude!)                */

  mdm_deinit();
  Fossil_Install();


  /* Redraw the status line after the fossil reinit, so that any FOSSIL     *
   * init messages don't bleed the status line onto the rest of the         *
   * screen.                                                                */
     
  Draw_StatusLine(STATUS_force);

  /*************************************************************************/
  /*                      Read .LOG file for protocols                     */
  /*************************************************************************/

  if (ctltype != CTL_none)  /* Read in log file */
    Read_External_Log(ctltype,parm,starttime,method);

  /*************************************************************************/
  /*                      Restart Video Display Junk                       */
  /*************************************************************************/

  /* We just ran an external program, so we have no idea what the          *
   * other program was doing with the user's screen location, so our       *
   * best guess is where the local cursor is on our screen...              */

  fossil_getxy(&current_line,&current_col);

  if (displaymode==VIDEO_ibm)
    VidGotoXY(current_col,current_line,TRUE);

  vbuf_flush();

  (void)display_slogan;

  if (returning)
    Display_File(0,NULL,returning);

  inchat=save_inchat;
  ChatRestoreStatus(&css);

  return erl;
#endif
}

#ifndef ORACLE

void pascal Close_Files(void)
{
  if (prm.flags & FLAG_close_sf)
  {
    LogClose();

    if (areasfile != -1)
      close(areasfile);

    if (areasidx != -1)
      close(areasidx);

    areasfile=-1;
    areasidx=-1;
  }
  else LogFlush();
}


void pascal Reopen_Files(void)
{
  if (prm.flags & FLAG_close_sf)
  {
    if (! LogOpen())
      quit(ERROR_critical);

    if ((areasfile=shopen(PRM(adat_name),O_RDONLY | O_BINARY))==-1)
    {
      logit(cantopen,PRM(adat_name));
      quit(ERROR_minfile);
    }

    if ((areasidx=shopen(PRM(aidx_name),O_RDONLY | O_BINARY))==-1)
    {
      logit(cantopen,PRM(aidx_name));
      quit(ERROR_minfile);
    }
  }
}



void pascal Barricade(int msg)
{
  FILE *barfile;

  int tries;

  char temp[BUFLEN],
       temp2[BUFLEN],
       temp3[BUFLEN];

  barricade_ok=FALSE;

  if ((barfile=shfopen(msg ? area.msgbar : area.filebar,fopen_read,O_RDONLY))==NULL)
  {
    logit(cf_barricade,msg ? area.msgbar : area.filebar);
    return;
  }

  while (fgets(temp2,BUFLEN,barfile) != NULL)
  {
    if (*temp2=='!')
    {
      getword(temp2+1,temp,ctl_delim,1);

      if (eqstri(fancy_str(Strip_Underscore(temp)),usr.name))
      {
        getword(temp2+1,temp,ctl_delim,2);
        barricade_priv=Parse_BarPriv(temp);
        fclose(barfile);
        return;
      }
      else if (eqstri(temp,all))
      {
        getword(temp2+1,temp,ctl_delim,2);

        if (*temp)
          barricade_priv=Parse_BarPriv(temp);
        else
        {
          barricade_priv=usr.priv;
          barricade_ok=TRUE;
        }

        fclose(barfile);
        return;
      }
    }
  }

  fclose(barfile);

  if (isnl(linebuf))
    Display_File(0,NULL,PRM(barricade));

  for (tries=1;tries <= 3;tries++)
  {
    OpusGetse(temp,'.',s(bar_access));
    display_line=display_col=1;

    fancy_str(temp);

    if (isnl(temp))
    {
      barricade_ok=FALSE;
      break;
    }

    if ((barfile=shfopen(msg ? area.msgbar : area.filebar,fopen_read,O_RDONLY))==NULL)
    {
      logit(cf_barricade,msg ? area.msgbar : area.filebar);
      return;
    }

    while (! feof(barfile))
    {
      fgets(temp2,BUFLEN,barfile);

      if (! feof(barfile))
      {
        getword(temp2,temp3,cmd_delim,1);

        if (*temp3 != '!' && eqstr(temp,fancy_str(temp3)))
        {
          getword(temp2,temp3,ctl_delim,2);
          barricade_priv=Parse_BarPriv(temp3);
          break;
        }
      }
    }

    if (barricade_priv==0x1234)
      break;

    if (! barricade_ok)
    {
      logit(s(log_bad_pwd),temp);

      Printf(s(wrong_pwd),tries);
      Putc('\n');

      if (tries==3)
      {
        logit(l_invalid_pwd);
        Puts(invalid_pwd);
        mdm_hangup();
      }
    }

    fclose(barfile);

    if (barricade_ok)
      break;
  }
}


void pascal Write_External_Ctlfile(int ctltype,char *parm,int method)
{
  FILE *ctl;

  char temp[PATHLEN],
       temp2[PATHLEN];

  int x;

  if (Form_Outside_Filename(parm,temp,method))
    return;

  strcat(temp,dotctl);

  if ((ctl=shfopen(Add_Task_Number(temp,temp2),fopen_write,O_WRONLY | O_CREAT | O_TRUNC))==NULL)
    logit(cantopen,fancy_str(temp2));
  else
  {
    if (local)
      fprintf(ctl,"Keyboard\n");
    else fprintf(ctl,"Port %u\nBaud %ld\n",port+1,baud);

    if (!local)
      fprintf(ctl,"Modem %x %lx %x %x %lx\n",port+1,baud,
              prm.handshake_mask /* 3 */,prm.carrier_mask,baud);

    fprintf(ctl,"Time %u\n",timeleft());

    if (log_name && *log_name)
      fprintf(ctl,"Log %s\n",log_name);

    fprintf(ctl,"Messages %s\n",area.msgpath);
    fprintf(ctl,"Uploads %s\n",area.uppath);
    fprintf(ctl,"Downloads %s\n",area.filepath);
    fprintf(ctl,"Help %sHlp\n",original_path);

    if (*area.filesbbs)
      fprintf(ctl,"FilesBbs %s\n",area.filesbbs);

    if (filenames[0])
    {
      if (ctltype==CTL_download)
      {
        for (x=0;x < fnames;x++)
          fprintf(ctl,"Send %s%s\n",area.filepath,filenames[x]);
      }
      else if (ctltype==CTL_upload)
      {
        for (x=0;x < fnames;x++)
          fprintf(ctl,"Get %s\n",filenames[x]);
      }
    }

    fclose(ctl);
  }
}



void pascal Read_External_Log(int ctltype,char *parm,long starttime,int method)
{
  FILE *log;

  char string[PATHLEN],
       temp[PATHLEN],
       save_tc,
       *s,
       *s2,
       gotafile;

  long bytes;

  gotafile=FALSE;

  if (Form_Outside_Filename(parm,temp,method))
    return;

  strcat(temp,dotlog);

  save_tc=do_timecheck;
  do_timecheck=FALSE;

  if ((log=shfopen(Add_Task_Number(temp,string),fopen_read,O_RDONLY))==NULL)
  {
    if (ctltype==CTL_upload || ctltype==CTL_download)
      logit(cantopen,fancy_str(string));
    /* Else do nothing! */
  }
  else
  {
    while (1)
    {
      fgets(temp,PATHLEN,log);
      Trim_Line(temp);

      if (! feof(log))
      {
        getword(temp,string,ctl_delim,1);

        if (!isnl(string))
        {
          if (eqstri(string,"Got"))
          {
            gotafile=TRUE;

            getword(temp,string,ctl_delim,3);
            bytes=atol(string);

            getword(temp,string,ctl_delim,2);

            /* Strip off path, if any */

            if (strpbrk(string,pdel_only))
              s2=strpbrk(string,pdel_only)+1;
            else s2=string;

            /* Remove any .BBS/.GBS extensions, if applicable */
            Check_Filename(area.uppath,s2);

            logit(s(log_ul),protocol_letter,area.uppath,s2);

            /* Get description */
            s=firstchar(temp,ctl_delim,4);

            Add_To_Upload_Log(area.uppath,s2,bytes);
            Add_Upload_Bytes(bytes);

            if (!s || (s && *s != '!'))
              Get_File_Description(s2,bytes,s);
          }
          else if (eqstri(string,"Sent"))
          {
            gotafile=TRUE;

            getword(temp,string,ctl_delim,2);

            logit(s(log_dl),protocol_letter,fancy_str(string));

            getword(temp,string,ctl_delim,3);

            bytes=atol(string);

            usr.dnld += (unsigned int)(bytes/1000L);
            usr.dnldl += (unsigned int)(bytes/1000L);
            bstats.total_dl += (dword)(bytes/1000L);
          }
        }
      }
      else break;
    }

    fclose(log);

    /* Now delete the .CTL and .LOG files */

    Form_Outside_Filename(parm,temp,method);
    strcpy(string,temp);
    strcat(string,dotlog);
    unlink(Add_Task_Number(string,string));

    strcpy(string,temp);
    strcat(string,dotctl);
    unlink(Add_Task_Number(string,string));
  }

  if (ctltype==CTL_upload || ctltype==CTL_download)
  {
    if (gotafile)
    {
      Puts(xfercomplete);
      Putc('\n');
    }
    else Puts(xferaborted);
  }

  if (ctltype==CTL_upload && gotafile)
  {
    starttime=Add_To_Time(((long)prm.ul_reward*(time(NULL)-starttime))/100L);

    if (starttime)
      Printf(time_added_for_ul,(long)(starttime/60L),(long)(starttime % 60L));
  }

  do_timecheck=save_tc;
}



void pascal Close_Chatlog(void)
{
  time_t ct;
  struct _stamp stamp;

  char temp[PATHLEN],
       p1[50],
       p2[50];

  TmDate_to_OpusDate(localtime(((ct=time(NULL)),&ct)),&stamp);

  strcpy(p1,Timestamp_Format(PRM(dateformat),&stamp,temp));
  strcpy(p2,Timestamp_Format(PRM(timeformat),&stamp,temp));

  fprintf(chatlog,s(chat_logend),p1,p2);

  fclose(chatlog);
  chatlog=NULL;
}



extern int pascal IsValidMsgArea(struct _area *a,int areanum);

void pascal Goodbye(void)
{
  struct _area ar;

  char string[PATHLEN],
       temp[PATHLEN];

  if (!local)
  {
    WhiteN();

    sprintf(string,"%sWHY_HU",PRM(misc_path));

    if (GetYnhAnswer(string,s(disconnect),0)==NO)
      return;

    WhiteN();

    Read_Area(&ar,Set_Area_Name(zero));
    ar.msgpriv=TWIT;

    /* Make sure there's a message path to leave comments in... */
    if (IsValidMsgArea(&ar,Set_Area_Name(zero)))
    {
      sprintf(string,"%sWHY_FB",PRM(misc_path));

      sprintf(temp,s(leave_msg),PRM(sysop));

      if (GetyNhAnswer(string,temp,0)==YES)
        Goodbye_Comment();
    }

    Display_File(0,NULL,PRM(byebye));

    Printf(s(bibi),usr.name);
    Delay(200);
  }

  mdm_hangup();
}





int pascal Parse_BarPriv(char *temp3)
{
  int barricade_priv,
      x;

  barricade_ok=TRUE;

  for (x=0;x < _PRIVS_NUM;x++)
    if (eqstri(temp3,_privs[x].name))
    {
      barricade_priv=_privs[x].priv;

      if (barricade_priv==HIDDEN)
        mdm_hangup();
      break;
    }

  if (x==_PRIVS_NUM)
  {
    if (eqstri(temp3,noaccess))
    {
      barricade_ok=FALSE;
      barricade_priv=0x1234;
    }
    else
    {
      logit(s(bar_priv_huh),temp3);
      barricade_priv=usr.priv;
    }
  }

  return barricade_priv;
}






/* Writes the RESTARxx.BBS file to the right directory. */

void pascal Write_Restart(char restart_type,char *restart_name,long restart_offset,int usrtime,int ctltype,char *parm,long starttime,struct _css *css)
{
  struct _restart rst;
  char temp[PATHLEN];
  int file;

  /* Now fill in the "restart" part of the structure */

  memset(&rst,'\0',sizeof(struct _restart));

  rst.rst_ver=RST_VER;
  rst.restart_type=restart_type;
  strcpy(rst.restart_name,restart_name);
  strcpy(rst.menupath,menupath);

  if (menuname)
    strcpy(rst.menuname,menuname);
  else strcpy(rst.menuname,main_menu);

  strcpy(rst.firstname,firstname);
  strcpy(rst.next_ldate,next_ldate);
  strcpy(rst.last_onexit,last_onexit);

  strcpy(rst.fix_menupath,fix_menupath);
  strcpy(rst.last_name,last_name);
  rst.mn_dirty=mn_dirty;

  rst.lastmenu=lastmenu;
  rst.local=local;
  rst.port=port;
  rst.snoop=snoop;
  rst.keyboard=keyboard;
  rst.protocol_letter=protocol_letter;
  rst.chatreq=chatreq;

  rst.timeon=timeon;
  rst.timeoff=timeoff;
  rst.timestart=timestart;
  rst.restart_offset=restart_offset;

  rst.usr_time=usrtime;
  rst.usernum=usernum;
  rst.ultoday=ultoday;
  rst.written_echomail=written_echomail;
  rst.written_matrix=written_matrix;
  rst.current_baud=current_baud;
  rst.baud=baud;
  rst.max_time=max_time;
  rst.laston=laston.msg_st;
  rst.area=area;

  rst.ctltype=ctltype;
  strcpy(rst.parm,parm);
  rst.starttime=starttime;

  memcpy(rst.echo_written_in,echo_written_in,sizeof(echo_written_in));

  rst.locked=locked;
  rst.locked2=locked2;

  rst.lockpriv=lockpriv;
  rst.lockpriv2=lockpriv2;

  rst.barricade_ok=barricade_ok;
  
  rst.css=*css;

  /* Make sure that we don't delete the RESTARxx.BBS file when exiting... */
  if (rst_offset==-1L)
    rst_offset=0L;

  sprintf(temp,restarxx_bbs,original_path,task_num);

  if ((file=sopen(temp,O_CREAT | O_TRUNC | O_WRONLY | O_BINARY,SH_DENYNONE,S_IREAD | S_IWRITE))==-1)
    logit(cantopen,fancy_str(temp));
  else
  {
    write(file,(char *)&rst,sizeof(struct _restart));
    close(file);
  }
}


static int pascal Form_Outside_Filename(char *in,char *out,int method)
{
  if (method==OUTSIDE_errorlevel)
  {
    getword(in,out," .",2);

    if (! *out)
      return 1;
  }
  else getword(in,out,".",1);

  return 0;
}


static void pascal Outside_Reread(int tr)
{
  char temp[PATHLEN];
  int ffile;

  if (task_num)
    sprintf(temp,lastusxx_bbs,original_path,task_num);
  else sprintf(temp,lastuser_bbs,original_path);

  if ((ffile=shopen(temp,O_RDONLY | O_BINARY))==-1)
    logit(cantopen,fancy_str(temp));
  else
  {
    read(ffile,(char *)&usr,sizeof(struct _usr));
    close(ffile);

    /* The 'timeremain' field is calculated as a relative value...  In     *
     * other words, the door program can only add or subtract minutes      *
     * from the time-remaining value, and Maximus will figure out the      *
     * relative inc/decrement by comparing the amount of time, as it was   *
     * written into the file by Max itself, to the time limit that was     *
     * read back in this function.  This works well, since the time limit  *
     * will continue normally if the program doesn't touch the             *
     * time-remaining field.  If the program wants to guarantee that a     *
     * specific number of minutes are left when the program exits, then    *
     * it must calculate the current time when the external program is     *
     * started, and add the value in usr.timeremain to it.  It should      *
     * then calculate the desired value for usr.timeremain, to get         *
     * the time-off to equal the required value.  (The usr.timeremain      *
     * field can be either decremented or incremented, and Max will        *
     * adjust appropriately.)                                              */

    timeoff += (usr.timeremaining-tr)*60L;

    Find_Class_Number();
  }
}

#endif /* !ORACLE */

